import pyproj
import numpy as np

def distance_between_trajectories(lon1, lat1, lon2, lat2):
    '''Calculate the distances [m] between two trajectories'''

    assert len(lon1) == len(lat1) == len(lat1) == len(lat2)
    geod = pyproj.Geod(ellps='WGS84')
    _azimuth_forward, _a2, distance = geod.inv(lon1, lat1, lon2, lat2)

    return distance

def distance_along_trajectory(lon, lat):
    '''Calculate the distances [m] between points along a trajectory'''

    geod = pyproj.Geod(ellps='WGS84')
    _azimuth_forward, _a2, distance = geod.inv(lon[1:], lat[1:], lon[0:-1], lat[0:-1])

    return distance

def liu_weissberg(lon_obs, lat_obs, lon_model, lat_model, tolerance_threshold=1):
    '''
    Calculate skill score from normalized cumulative seperation distance. Liu and Weisberg 2011.

    Returns:

        Skill score between 0. and 1.
    '''

    lon_obs = np.array(lon_obs)
    lat_obs = np.array(lat_obs)
    lon_model = np.array(lon_model)
    lat_model = np.array(lat_model)
    d = distance_between_trajectories(lon_obs, lat_obs, lon_model, lat_model)
    l = distance_along_trajectory(lon_obs, lat_obs)
    s = np.nansum(d) / np.nansum(np.nancumsum(l))
    if tolerance_threshold==0:
        skillscore = 0
    else:
        skillscore = max(0, 1 - s/tolerance_threshold)

    return skillscore




"""
The OMB data decoder; slightly curated from:

https://github.com/jerabaul29/OpenMetBuoy-v2021a/tree/main/legacy_firmware/decoder

For issues related to the decoder, please discuss on the OMB repository:

https://github.com/jerabaul29/OpenMetBuoy-v2021a/issues
"""

import binascii
import struct
import datetime
from dataclasses import dataclass
import math
import numpy as np
import scipy.signal as signal
import logging

logger = logging.getLogger(__name__)

#--------------------------------------------------------------------------------
# a few module constants

_BD_VERSION_NBR = "2.1"

####################
# derived properties of the GNSS packets
# 14; 1 byte start, posix, 2 longs, 1 byte end
_BD_GNSS_PACKET_LENGTH = 14

####################
# properties of the wave spectra packets with 2048 FFT_LEN
_BD_YWAVE_PACKET_MIN_BIN = 9
_BD_YWAVE_PACKET_MAX_BIN = 64
_BD_YWAVE_PACKET_SCALER = 65000
_BD_YWAVE_PACKET_SAMPLING_FREQ_HZ = 10.0
_BD_YWAVE_PACKET_NBR_SAMPLES_PER_SEGMENT = 2**11

# derived properties of the wave packer
_BD_YWAVE_NBR_BINS = _BD_YWAVE_PACKET_MAX_BIN - _BD_YWAVE_PACKET_MIN_BIN
_BD_YWAVE_PACKET_FRQ_RES = _BD_YWAVE_PACKET_SAMPLING_FREQ_HZ / _BD_YWAVE_PACKET_NBR_SAMPLES_PER_SEGMENT

# 138 in total length; 1 byte start, posix, int, 4 floats, uint36 array, 2 bytes for alignment, 1 byte end
_BD_YWAVE_PACKET_LENGTH = \
    1 + 4 + 4 + 4*4 + _BD_YWAVE_NBR_BINS * 2 + 2 + 1
LENGTH_FROM_SERIAL_OUTPUT = 138
assert _BD_YWAVE_PACKET_LENGTH == LENGTH_FROM_SERIAL_OUTPUT, "the arduino printout indicates that wave packets have length {}".format(
    LENGTH_FROM_SERIAL_OUTPUT)

####################
# properties of the thermistors packets
_BD_THERM_MSG_FIXED_LENGTH = 3  # start byte, byte metadata nbr packets, byte end
_BD_THERM_MSG_NBR_THERMISTORS = 6
_BD_THERM_PACKET_NBR_BYTES_PER_THERMISTOR = 3
_BD_THERM_PACKET_LENGTH = 1 + 1 * 4 + _BD_THERM_PACKET_NBR_BYTES_PER_THERMISTOR * _BD_THERM_MSG_NBR_THERMISTORS * 1 + 6 * 1
# these are from the params of the tracker
_BD_THERM_ROLL_FLOAT_TO_INT8_FACTOR = 0.7
_BD_THERM_PITCH_FLOAT_TO_INT8_FACTOR = 1.4
# factor for converting bin 12 bits signed int to temperature; from thermistor datasheet
_BD_THERM_12BITS_TO_FLOAT_TEMPERATURE_FACTOR = 1.0 / 16.0

#--------------------------------------------------------------------------------
# misc


def get_version():
    return _BD_VERSION_NBR


#--------------------------------------------------------------------------------
# helper functions for unpacking binary data


def byte_to_char(crrt_byte):
    return (chr(crrt_byte))


def one_byte_to_int(crrt_byte):
    return (struct.unpack('B', bytes(crrt_byte))[0])


def one_byte_to_signed_int(crrt_byte):
    return (struct.unpack('b', bytes(crrt_byte))[0])


def four_bytes_to_long(crrt_four_bytes):
    res = struct.unpack('<l', bytes(crrt_four_bytes))
    return res[0]


def four_bytes_to_int(crrt_four_bytes):
    res = struct.unpack('<i', bytes(crrt_four_bytes))
    return res[0]


def four_bytes_to_unsignedint(crrt_four_bytes):
    res = struct.unpack('<I', bytes(crrt_four_bytes))
    return res[0]


def four_bytes_to_float(crrt_four_bytes):
    res = struct.unpack('<f', bytes(crrt_four_bytes))
    return res[0]


#--------------------------------------------------------------------------------
# custom data classes to store data packets


@dataclass
class Spectral_Moments:
    m0: float
    m2: float
    m4: float


@dataclass
class GNSS_Packet:
    datetime_fix: datetime.datetime
    datetime_posix: int
    latitude: float
    longitude: float
    is_valid: bool


@dataclass
class GNSS_Metadata:
    nbr_gnss_fixes: int


@dataclass
class Waves_Packet:
    # the decoded part: this is just decoding / textbook operations
    datetime_fix: datetime.datetime
    datetime_posix: int
    spectrum_number: int
    Hs: float
    Tz: float
    Tc: float
    _array_max_value: float
    _array_uint16: float
    list_frequencies: list
    list_acceleration_energies: list
    frequency_resolution: float
    list_elevation_energies: list
    wave_spectral_moments: Spectral_Moments
    is_valid: bool
    # the processed part: this relies on some more advanced operations
    processed_list_frequencies: list
    processed_list_elevation_energies: list
    processed_wave_spectral_moments: Spectral_Moments
    processed_Hs: float
    processed_Tz: float
    processed_Tc: float
    low_frequency_index_cutoff: int
    # processed_quality_index: str  # todo: add a quality explanation string


@dataclass
class Waves_Metadata:
    None


@dataclass
class Thermistors_Reading:
    mean_temperature: float
    range_temperature: float
    probe_id: int


@dataclass
class Thermistors_Packet:
    datetime_packet: datetime.datetime
    datetime_posix: int
    thermistors_readings: list
    mean_pitch: float
    min_pitch: float
    max_pitch: float
    mean_roll: float
    min_roll: float
    max_roll: float


@dataclass
class Thermistors_Metadata:
    nbr_thermistors_measurements: int


# --------------------------------------------------------------------------------
# data quality and processing utils


def find_low_frequency_cutoff(list_frequencies, list_elevation_energies):
    """Find the low frequency cutoff to avoid double integration noise contamination
    for IMU measurements of waves, similar to what is discussed in Fig. 7 of
    https://www.mdpi.com/2076-3263/12/3/110 .

    Inputs:
        list_frequencies: the list of frequencies at which spectrum is provided
        list_elevation_energies: the wave spectrum elevation energies at the corresponding
            frequencies
    Output:
        index_low_frequency_cutoff: the index to use to cut off low frequency double
            integration noise contamination; i.e., use only indexes > to the output
            to consider the valid part of the spectrum.
    """

    assert isinstance(list_frequencies, list)
    assert isinstance(list_elevation_energies, list)
    assert len(list_frequencies) == len(list_elevation_energies)

    # findpeaks:
    # we look for minima (peaks looks for maxima, so -)
    # we want peaks that are local peaks in their neighborhood and not just "super local minima": require distance = 3
    # we want only peaks that are clear enough: normalize the spectrum so that the maximum is always 1.0, and want a prominence of at least 0.05
    normalized_spectrum = -np.array(list_elevation_energies) / np.max(
        list_elevation_energies)
    peaks_output = signal.find_peaks(normalized_spectrum,
                                     distance=3,
                                     prominence=0.05)

    peaks = list(peaks_output[0])
    if len(peaks) == 0:
        peaks = [0]

    # isolate the peak we want
    # we are only interested in the first minimum, and it has to be low enough that it does correspond to a low freq. threshold
    first_peak = peaks[0]
    if list_frequencies[
            first_peak] > 0.10:  # the 0.1 value is somewhat arbitrary, but in practice we always have some energy at least there or before
        first_peak = 0  # we keep all indexes if there was no clear minimum on the left (ie no low energy noise)

    # we do not want to flag out valid parts of the spectrum when the spectrum is "really clean"
    # in cases where the spectrum is "really clean", can happen that the first minimum is a local minimum after the first valid peak
    # detect these cases and set the full spectrum as valid then
    if (list_elevation_energies[first_peak] >
        ((list_elevation_energies[0] + list_elevation_energies[1]) / 2.0)):
        first_peak = 0

    index_low_frequency_cutoff = first_peak

    return index_low_frequency_cutoff


# --------------------------------------------------------------------------------
# packets and messages decoding


def hex_to_bin_message(hex_string_message, print_info=False):
    bin_msg = binascii.unhexlify(hex_string_message)
    return bin_msg


def message_kind(bin_msg):
    first_char = byte_to_char(bin_msg[0])
    valid_first_chars = ["G", "Y", "T"]
    assert first_char in valid_first_chars, "unknown first_char message kind: got {}, valids are {}".format(
        first_char, valid_first_chars)
    return first_char


def decode_gnss_packet(bin_packet,
                       print_decoded=False,
                       print_debug_information=False):
    assert len(
        bin_packet
    ) == _BD_GNSS_PACKET_LENGTH, "GNSS packets with start and end byte have 14 bytes, got {} bytes".format(
        len(bin_packet))

    char_first_byte = byte_to_char(bin_packet[0])

    assert char_first_byte == 'F', "GNSS packets must start with a 'F', got {}".format(
        char_first_byte)

    posix_timestamp_fix = four_bytes_to_long(bin_packet[1:5])
    # print(posix_timestamp_fix)
    datetime_fix = datetime.datetime.utcfromtimestamp(posix_timestamp_fix)

    latitude_long = four_bytes_to_long(bin_packet[5:9])
    latitude = latitude_long / 1.0e7

    longitude_long = four_bytes_to_long(bin_packet[9:13])
    longitude = longitude_long / 1.0e7

    if print_decoded:
        print("-------------------- decoded GNSS packet ---------------------")
        print("fix at posix {}, i.e. {}".format(posix_timestamp_fix,
                                                datetime_fix))
        print("latitude {}, i.e. {}".format(latitude_long, latitude))
        print("longitude {}, i.e. {}".format(longitude_long, longitude))
        print("--------------------------------------------------------------")

    char_next_byte = byte_to_char(bin_packet[13])

    assert char_next_byte == 'E' or char_next_byte == 'F', "either end ('E') or fix ('F') expected at the end, got {}".format(
        char_next_byte)

    decoded_packet = GNSS_Packet(datetime_fix=datetime_fix,
                                 datetime_posix=posix_timestamp_fix,
                                 latitude=latitude,
                                 longitude=longitude,
                                 is_valid=True)

    return decoded_packet


def decode_gnss_message(bin_msg,
                        print_decoded=True,
                        print_debug_information=False):
    if print_decoded:
        print(
            "----------------------- START DECODE GNSS MESSAGE -----------------------"
        )

    assert message_kind(bin_msg) == 'G'
    expected_message_length = int(1 +
                                  (len(bin_msg) - 2 - _BD_GNSS_PACKET_LENGTH) /
                                  (_BD_GNSS_PACKET_LENGTH - 1))

    if print_decoded:
        print("expected number of packets based on message length: {}".format(
            expected_message_length))

    nbr_gnss_fixes = one_byte_to_int(bin_msg[1:2])
    message_metadata = GNSS_Metadata(nbr_gnss_fixes=nbr_gnss_fixes)

    if print_decoded:
        print("number of fixes since boot at message creation: {}".format(
            nbr_gnss_fixes))

    crrt_packet_start = 2
    list_decoded_packets = []

    while True:
        crrt_byte_start = byte_to_char(bin_msg[crrt_packet_start])
        assert crrt_byte_start == "F"

        crrt_decoded_packet = decode_gnss_packet(
            bin_msg[crrt_packet_start:crrt_packet_start +
                    _BD_GNSS_PACKET_LENGTH],
            print_decoded=print_decoded,
            print_debug_information=print_debug_information)
        list_decoded_packets.append(crrt_decoded_packet)

        trailing_char = byte_to_char(bin_msg[crrt_packet_start +
                                             _BD_GNSS_PACKET_LENGTH - 1])
        assert trailing_char in ["E", "F"]
        if trailing_char == "E":
            break
        else:
            crrt_packet_start += _BD_GNSS_PACKET_LENGTH - 1

    assert expected_message_length == len(list_decoded_packets)

    if print_decoded:
        print(
            "----------------------- DONE DECODE GNSS MESSAGE -----------------------"
        )

    return message_metadata, list_decoded_packets


def decode_ywave_packet(bin_packet,
                        print_decoded=False,
                        print_debug_information=False):
    assert len(bin_packet) == _BD_YWAVE_PACKET_LENGTH

    char_first_byte = byte_to_char(bin_packet[0])
    assert char_first_byte == "Y"

    char_last_byte = byte_to_char(bin_packet[-1])
    assert char_last_byte == "E"

    crrt_start_data_field = 1

    posix_timestamp = four_bytes_to_long(
        bin_packet[crrt_start_data_field:crrt_start_data_field + 4])
    crrt_start_data_field += 4
    datetime_packet = datetime.datetime.utcfromtimestamp(posix_timestamp)

    spectrum_number = four_bytes_to_int(
        bin_packet[crrt_start_data_field:crrt_start_data_field + 4])
    crrt_start_data_field += 4

    Hs = four_bytes_to_float(
        bin_packet[crrt_start_data_field:crrt_start_data_field + 4])
    crrt_start_data_field += 4

    Tz = 1.0 / four_bytes_to_float(
        bin_packet[crrt_start_data_field:crrt_start_data_field + 4])
    crrt_start_data_field += 4

    Tc = 1.0 / four_bytes_to_float(
        bin_packet[crrt_start_data_field:crrt_start_data_field + 4])
    crrt_start_data_field += 4

    _array_max_value = four_bytes_to_float(
        bin_packet[crrt_start_data_field:crrt_start_data_field + 4])
    crrt_start_data_field += 4

    nbr_bytes_uint16_array = _BD_YWAVE_NBR_BINS * 2
    _array_uint16 = struct.unpack(
        '<' + _BD_YWAVE_NBR_BINS * "H",
        bin_packet[crrt_start_data_field:crrt_start_data_field +
                   nbr_bytes_uint16_array])
    crrt_start_data_field += nbr_bytes_uint16_array

    assert crrt_start_data_field + 1 + 2 == _BD_YWAVE_PACKET_LENGTH  # the +2 is due to struct alignment issues

    if Hs > 1e-5:
        is_valid = True
    else:
        is_valid = False
        logger.warning(f"got a wave spectrum with corresponding Hs<1e-5; this likely indicates an instrument that landed onshore or an IMU malfunction; setting corresponding spectrum is_valid to False")

    list_frequencies = []
    list_acceleration_energies = []

    for ind, crrt_uint16 in enumerate(_array_uint16):
        list_frequencies.append(
            (_BD_YWAVE_PACKET_MIN_BIN + ind) * _BD_YWAVE_PACKET_FRQ_RES)
        list_acceleration_energies.append(crrt_uint16 * _array_max_value /
                                          _BD_YWAVE_PACKET_SCALER)

    list_omega = [2.0 * math.pi * crrt_freq for crrt_freq in list_frequencies]
    list_omega_4 = [math.pow(crrt_omega, 4) for crrt_omega in list_omega]
    list_elevation_energies = [
        crrt_acceleration_energy / crrt_omega_4
        for (crrt_acceleration_energy,
             crrt_omega_4) in zip(list_acceleration_energies, list_omega_4)
    ]

    def compute_spectral_moment(list_frequencies, list_elevation_energies,
                                order):
        list_to_integrate = [
            math.pow(crrt_freq, order) * crrt_energy
            for (crrt_freq,
                 crrt_energy) in zip(list_frequencies, list_elevation_energies)
        ]

        moment = np.trapz(list_to_integrate, list_frequencies)

        return moment

    m0 = compute_spectral_moment(list_frequencies, list_elevation_energies, 0)
    m2 = compute_spectral_moment(list_frequencies, list_elevation_energies, 2)
    m4 = compute_spectral_moment(list_frequencies, list_elevation_energies, 4)

    spectral_moments = Spectral_Moments(m0, m2, m4)

    # add the post processed part

    low_frequency_index_cutoff = find_low_frequency_cutoff(
        list_frequencies, list_elevation_energies)

    processed_list_frequencies = list_frequencies[low_frequency_index_cutoff:]
    processed_list_elevation_energies = list_elevation_energies[
        low_frequency_index_cutoff:]

    processed_m0 = compute_spectral_moment(processed_list_frequencies,
                                           processed_list_elevation_energies,
                                           0)
    processed_m2 = compute_spectral_moment(processed_list_frequencies,
                                           processed_list_elevation_energies,
                                           2)
    processed_m4 = compute_spectral_moment(processed_list_frequencies,
                                           processed_list_elevation_energies,
                                           4)

    processed_wave_spectral_moments = Spectral_Moments(
        processed_m0,
        processed_m2,
        processed_m4,
    )

    processed_Hs = 4 * math.sqrt(processed_m0)
    processed_Tz = 1.0 / math.sqrt(processed_m2 / processed_m0)
    processed_Tc = 1.0 / math.sqrt(processed_m4 / processed_m2)

    processed_list_frequencies = list_frequencies
    processed_list_elevation_energies = low_frequency_index_cutoff * [
        math.nan
    ] + processed_list_elevation_energies

    decoded_packet = Waves_Packet(
        datetime_packet,
        posix_timestamp,
        spectrum_number,
        Hs,
        Tz,
        Tc,
        _array_max_value,
        _array_uint16,
        list_frequencies,
        list_acceleration_energies,
        _BD_YWAVE_PACKET_FRQ_RES,
        list_elevation_energies,
        spectral_moments,
        is_valid,
        processed_list_frequencies,
        processed_list_elevation_energies,
        processed_wave_spectral_moments,
        processed_Hs,
        processed_Tz,
        processed_Tc,
        low_frequency_index_cutoff,
    )

    return decoded_packet


def decode_ywave_message(bin_msg,
                         print_decoded=True,
                         print_debug_information=False):
    if print_decoded:
        print(
            "----------------------- START DECODE YWAVES MESSAGE -----------------------"
        )

    assert message_kind(bin_msg) == "Y"
    assert byte_to_char(bin_msg[_BD_YWAVE_PACKET_LENGTH - 1]) == "E"

    message_metadata = Waves_Metadata()

    list_decoded_packets = []
    crrt_packet = decode_ywave_packet(
        bin_msg,
        print_decoded=print_decoded,
        print_debug_information=print_debug_information)
    list_decoded_packets.append(crrt_packet)

    if print_decoded:
        print(
            "----------------------- DONE DECODE GNSS MESSAGE -----------------------"
        )

    return message_metadata, list_decoded_packets


def decode_thermistor_reading(crrt_thermistor_bin,
                              print_debug_information=False):

    # quite a bit of tweaking...
    # this is the inverse operation of the binary encoding done in the thermistor manager C++ code

    # TODO: actually, this can be made with clear syntax a la n & 0xffffffff too, fixme

    id_6_bits = one_byte_to_int(crrt_thermistor_bin[0:1]) // 4

    reading_2_higher_bits = one_byte_to_int(crrt_thermistor_bin[0:1]) % 4
    reading_2_higher_bits_lower = reading_2_higher_bits % 2
    reading_2_higher_bits_higher = (reading_2_higher_bits -
                                    reading_2_higher_bits_lower) // 2
    reading_8_middle_bits = one_byte_to_int(crrt_thermistor_bin[1:2])
    reading_2_lower_bits = one_byte_to_int(crrt_thermistor_bin[2:3]) // 64

    reading_reconstructed_bin = reading_2_lower_bits + (
        2**2) * reading_8_middle_bits + (2**10) * reading_2_higher_bits_lower
    if reading_2_higher_bits_higher:
        reading_reconstructed_bin = reading_reconstructed_bin - 2**11 - 1

    range_6_bits_bin = one_byte_to_int(crrt_thermistor_bin[2:3]) % 64

    reading_reconstructed = reading_reconstructed_bin * _BD_THERM_12BITS_TO_FLOAT_TEMPERATURE_FACTOR
    range_temperature = range_6_bits_bin * _BD_THERM_12BITS_TO_FLOAT_TEMPERATURE_FACTOR

    crrt_thermistor_reading = Thermistors_Reading(
        mean_temperature=reading_reconstructed,
        range_temperature=range_temperature,
        probe_id=id_6_bits)

    return crrt_thermistor_reading


def decode_thermistors_packet(bin_packet,
                              print_decoded=False,
                              print_debug_information=False):
    if print_debug_information:
        print("----- START DECODE THERM PACKET -----")

    assert len(bin_packet) == _BD_THERM_PACKET_LENGTH

    char_first_byte = byte_to_char(bin_packet[0])
    assert char_first_byte == "P"

    crrt_start_field = 1

    posix_timestamp = four_bytes_to_long(
        bin_packet[crrt_start_field:crrt_start_field + 4])
    datetime_packet = datetime.datetime.utcfromtimestamp(posix_timestamp)
    crrt_start_field += 4

    list_thermistors_readings = []

    for crrt_thermistor in range(_BD_THERM_MSG_NBR_THERMISTORS):
        crrt_thermistor_bin = bin_packet[crrt_start_field:crrt_start_field + 3]
        crrt_start_field += 3
        crrt_thermistor_reading = decode_thermistor_reading(
            crrt_thermistor_bin,
            print_debug_information=print_debug_information)
        assert isinstance(crrt_thermistor_reading, Thermistors_Reading)
        list_thermistors_readings.append(crrt_thermistor_reading)

    mean_pitch_bin = one_byte_to_signed_int(
        bin_packet[crrt_start_field:crrt_start_field + 1])
    crrt_start_field += 1

    mean_roll_bin = one_byte_to_signed_int(
        bin_packet[crrt_start_field:crrt_start_field + 1])
    crrt_start_field += 1

    min_pitch_bin = one_byte_to_signed_int(
        bin_packet[crrt_start_field:crrt_start_field + 1])
    crrt_start_field += 1

    max_pitch_bin = one_byte_to_signed_int(
        bin_packet[crrt_start_field:crrt_start_field + 1])
    crrt_start_field += 1

    min_roll_bin = one_byte_to_signed_int(
        bin_packet[crrt_start_field:crrt_start_field + 1])
    crrt_start_field += 1

    max_roll_bin = one_byte_to_signed_int(
        bin_packet[crrt_start_field:crrt_start_field + 1])
    crrt_start_field += 1

    assert crrt_start_field == _BD_THERM_PACKET_LENGTH

    crrt_thermistor_packet = Thermistors_Packet(
        datetime_packet=datetime_packet,
        datetime_posix=posix_timestamp,
        thermistors_readings=list_thermistors_readings,
        mean_pitch=mean_pitch_bin / _BD_THERM_PITCH_FLOAT_TO_INT8_FACTOR,
        min_pitch=min_pitch_bin / _BD_THERM_PITCH_FLOAT_TO_INT8_FACTOR,
        max_pitch=max_pitch_bin / _BD_THERM_PITCH_FLOAT_TO_INT8_FACTOR,
        mean_roll=mean_roll_bin / _BD_THERM_ROLL_FLOAT_TO_INT8_FACTOR,
        min_roll=min_roll_bin / _BD_THERM_ROLL_FLOAT_TO_INT8_FACTOR,
        max_roll=max_roll_bin / _BD_THERM_ROLL_FLOAT_TO_INT8_FACTOR)

    if print_debug_information:
        print("----- DONE DECODE THERM PACKET -----")

    return crrt_thermistor_packet


def decode_thermistors_message(bin_msg,
                               print_decoded=False,
                               print_debug_information=False):
    if print_decoded:
        print(
            "----------------------- START DECODE THERMISTORS MESSAGE -----------------------"
        )

    assert message_kind(bin_msg) == "T"
    assert byte_to_char(bin_msg[-1]) == "E"

    if (print_debug_information):
        print("received message of length: {}".format(len(bin_msg)))

    expected_message_length = int(
        (len(bin_msg) - _BD_THERM_MSG_FIXED_LENGTH) / _BD_THERM_PACKET_LENGTH)
    assert expected_message_length * _BD_THERM_PACKET_LENGTH + _BD_THERM_MSG_FIXED_LENGTH == len(
        bin_msg)

    nbr_thermistors_measurements = one_byte_to_int(bin_msg[1:2])
    message_metadata = Thermistors_Metadata(
        nbr_thermistors_measurements=nbr_thermistors_measurements)

    if print_decoded:
        print(message_metadata)

    crrt_packet_start = 2
    list_decoded_packets = []

    while True:
        if print_decoded:
            print("----- START PACKET -----")

        crrt_byte_start = byte_to_char(bin_msg[crrt_packet_start])
        assert crrt_byte_start == "P"

        # decode
        crrt_decoded_packet = decode_thermistors_packet(
            bin_msg[crrt_packet_start:crrt_packet_start +
                    _BD_THERM_PACKET_LENGTH],
            print_decoded=print_decoded,
            print_debug_information=print_debug_information)
        list_decoded_packets.append(crrt_decoded_packet)

        trailing_char = byte_to_char(bin_msg[crrt_packet_start +
                                             _BD_THERM_PACKET_LENGTH])
        assert trailing_char in ["P", "E"]
        if trailing_char == "E":
            break
        else:
            crrt_packet_start += _BD_THERM_PACKET_LENGTH

        if print_decoded:
            print("----- END PACKET -----")

    assert expected_message_length == len(list_decoded_packets)

    if print_decoded:
        print(
            "----------------------- DONE DECODE THERMISTORS MESSAGE -----------------------"
        )

    return message_metadata, list_decoded_packets


def decode_message(hex_string_message,
                   print_decoded=True,
                   print_debug_information=False,
                   dict_wave_packet_params=None):
    # NOTE: this is pushing the decoder further than it was initially thought...
    # some users change the default size of the transmitted spectra; as the logics of the packets formats were initially done
    # as global variables, this implies doing some hacking for allowing to change this on the fly
    # the on the fly globals editing below is dangerous and bad; that would need refactor in the future
    # this decoder has outgrown itself :)
    if dict_wave_packet_params is not None:
        global _BD_YWAVE_PACKET_MIN_BIN
        global _BD_YWAVE_PACKET_MAX_BIN
        global _BD_YWAVE_NBR_BINS
        global _BD_YWAVE_PACKET_LENGTH
        global LENGTH_FROM_SERIAL_OUTPUT

        old_BD_YWAVE_PACKET_MIN_BIN = _BD_YWAVE_PACKET_MIN_BIN
        old_BD_YWAVE_PACKET_MAX_BIN = _BD_YWAVE_PACKET_MAX_BIN
        old_BD_YWAVE_NBR_BINS = _BD_YWAVE_NBR_BINS
        old_BD_YWAVE_PACKET_LENGTH = _BD_YWAVE_PACKET_LENGTH
        oldLENGTH_FROM_SERIAL_OUTPUT = LENGTH_FROM_SERIAL_OUTPUT

        _BD_YWAVE_PACKET_MIN_BIN = dict_wave_packet_params[
            "_BD_YWAVE_PACKET_MIN_BIN"]
        _BD_YWAVE_PACKET_MAX_BIN = dict_wave_packet_params[
            "_BD_YWAVE_PACKET_MAX_BIN"]
        _BD_YWAVE_NBR_BINS = _BD_YWAVE_PACKET_MAX_BIN - _BD_YWAVE_PACKET_MIN_BIN
        _BD_YWAVE_PACKET_LENGTH = \
            1 + 4 + 4 + 4*4 + _BD_YWAVE_NBR_BINS * 2 + 2 + 1
        LENGTH_FROM_SERIAL_OUTPUT = dict_wave_packet_params[
            "LENGTH_FROM_SERIAL_OUTPUT"]

        assert _BD_YWAVE_PACKET_LENGTH == LENGTH_FROM_SERIAL_OUTPUT, "the arduino printout indicates that wave packets have length {}".format(
            LENGTH_FROM_SERIAL_OUTPUT)

    bin_msg = hex_to_bin_message(hex_string_message)

    kind = message_kind(bin_msg)

    if kind == "G":
        message_metadata, list_decoded_packets = decode_gnss_message(
            bin_msg,
            print_decoded=print_decoded,
            print_debug_information=print_debug_information)
    elif kind == "Y":
        message_metadata, list_decoded_packets = decode_ywave_message(
            bin_msg,
            print_decoded=print_decoded,
            print_debug_information=print_debug_information)
    elif kind == "T":
        message_metadata, list_decoded_packets = decode_thermistors_message(
            bin_msg,
            print_decoded=print_decoded,
            print_debug_information=print_debug_information)
    else:
        raise RuntimeError("Unknown message kind: {}".format(kind))

    if dict_wave_packet_params is not None:
        _BD_YWAVE_PACKET_MIN_BIN = old_BD_YWAVE_PACKET_MIN_BIN
        _BD_YWAVE_PACKET_MAX_BIN = old_BD_YWAVE_PACKET_MAX_BIN
        _BD_YWAVE_NBR_BINS = old_BD_YWAVE_NBR_BINS
        _BD_YWAVE_PACKET_LENGTH = old_BD_YWAVE_PACKET_LENGTH
        LENGTH_FROM_SERIAL_OUTPUT = oldLENGTH_FROM_SERIAL_OUTPUT

    return (kind, message_metadata, list_decoded_packets)



import xarray as xr
import numpy as np
import pandas as pd
import logging

from .traj import Traj, __require_obsdim__

logger = logging.getLogger(__name__)


class Traj2d(Traj):
    """
    A unstructured dataset, where each trajectory may have observations at different times. Typically from a collection of drifters.
    """

    def __init__(self, ds, obsdim, timedim):
        super().__init__(ds, obsdim, timedim)

    def timestep(self, average=np.nanmedian):
        """
        Return median time step between observations in seconds.
        """
        td = np.diff(self.ds.time, axis=1) / np.timedelta64(1, 's')
        td = average(td)
        return td

    def time_to_next(self):
        """Returns time from one position to the next

           Returned datatype is np.timedelta64
           Last time is repeated for last position (which has no next position)
        """
        time = self.ds.time
        lenobs = self.ds.sizes['obs']
        td = time.isel(obs=slice(1, lenobs)) - time.isel(
            obs=slice(0, lenobs - 1))
        td = xr.concat((td, td.isel(obs=-1)),
                       dim='obs')  # repeating last time step
        return td

    def is_1d(self):
        return False

    def is_2d(self):
        return True

    def insert_nan_where(self, condition):
        """Insert NaN-values in trajectories after given positions, shifting rest of trajectory"""

        index_of_last = self.index_of_last()
        num_inserts = condition.sum(dim='obs')
        max_obs = (index_of_last + num_inserts).max().values

        # Create new empty dataset with extended obs dimension
        trajcoord = range(self.ds.sizes['trajectory'])
        nd = xr.Dataset(
            coords={
                'trajectory':
                (["trajectory"], range(self.ds.sizes['trajectory'])),
                'obs': (['obs'], range(max_obs))  # Longest trajectory
            },
            attrs=self.ds.attrs)

        # Add extended variables
        for varname, var in self.ds.data_vars.items():
            if 'obs' not in var.dims:
                nd[varname] = var
                continue
            # Create empty dataarray to hold interpolated values for given variable
            da = xr.DataArray(
                data=np.zeros(tuple(nd.sizes[di] for di in nd.dims)) * np.nan,
                dims=nd.dims,
                attrs=var.attrs,
            )

            for t in range(
                    self.ds.sizes['trajectory']):  # loop over trajectories
                numins = num_inserts[t]
                olddata = var.isel(trajectory=t).values
                wh = np.argwhere(condition.isel(trajectory=t).values) + 1
                if len(wh) == 0:
                    newdata = olddata
                else:
                    insert_indices = np.concatenate(wh)
                    s = np.split(olddata, insert_indices)

                    if np.issubdtype(var.dtype, np.datetime64):
                        na = np.atleast_1d(np.datetime64("NaT"))
                    else:
                        na = np.atleast_1d(np.nan)
                    newdata = np.concatenate(
                        [np.concatenate((ss, na)) for ss in s])

                newdata = newdata[slice(0, max_obs -
                                        1)]  # truncating, should be checked
                da[{'trajectory': t, 'obs': slice(0, len(newdata))}] = newdata

            nd[varname] = da.astype(var.dtype)

        nd = nd.drop_vars(('obs', 'trajectory'))  # Remove coordinates

        return nd

    def drop_where(self, condition):
        """Remove positions where condition is True, shifting rest of trajectory"""

        trajs = []
        newlen = 0
        for i in range(self.ds.sizes['trajectory']):
            new = self.ds.isel(trajectory=i).drop_sel(obs=np.where(
                condition.isel(
                    trajectory=i))[0])  # Dropping from given trajectory
            newlen = max(newlen, new.sizes['obs'])
            trajs.append(new)

        # Ensure all trajectories have equal length, by padding with NaN at end
        trajs = [
            t.pad(pad_width={'obs': (0, newlen - t.sizes['obs'])})
            for t in trajs
        ]

        return xr.concat(trajs, dim='trajectory')

    @__require_obsdim__
    def condense_obs(self):
        """
        Move all observations to the first index, so that the observation
        dimension is reduced to a minimum. When creating ragged arrays the
        observations from consecutive trajectories start at the observation
        index after the previous, causing a very long observation dimension.

        Original:

        .............. Observations --->
        trajectory 1: | t01 | t02 | t03 | t04 | t05 | nan | nan | nan | nan |
        trajectory 2: | nan | nan | nan | nan | nan | t11 | t12 | t13 | t14 |

        After condensing:

        .............. Observations --->
        trajectory 1: | t01 | t02 | t03 | t04 | t05 |
        trajectory 2: | t11 | t12 | t13 | t14 | nan |

        Returns:

            A new Dataset with observations condensed.
        """

        on = self.ds.sizes[self.obsdim]
        logger.debug(f'Condensing {on} observations.')

        ds = self.ds.copy(deep=True)

        # The observation coordinate will be re-written
        ds = ds.drop_vars([self.obsdim])

        assert self.obsdim in ds[
            self.timedim].dims, "observation not a coordinate of time variable"

        # Move all observations for each trajectory to starting row
        maxN = 0
        for ti in range(len(ds.trajectory)):
            obsvars = [
                var for var in ds.variables if self.obsdim in ds[var].dims
            ]
            iv = np.full(on, False)
            for var in obsvars:
                ivv = ~pd.isnull(
                    ds[self.timedim][ti, :])  # valid times in this trajectory.
                iv = np.logical_or(iv, ivv)

            N = np.count_nonzero(iv)
            maxN = max(N, maxN)
            # logger.debug(f'Condensing trajectory {ti=}, observations: {N}..')

            if N == 0:
                logger.error(f'No valid observations in trajectory {ti}.')
                continue

            for var in obsvars:
                # logger.debug(f'Condensing variable {var}..')
                n = np.count_nonzero(~pd.isnull(ds[var][ti, :]))
                assert n <= N, f"Unexpected number of observations in trajectory for {ti=}, {var}: {n} > {N}."

                ds[var][ti, :N] = ds[var][ti, iv]
                ds[var][ti, N:] = np.nan

                # assert (~np.isnan(ds[var][ti, :N])).all(
                # ), "Varying number of valid observations within same trajectory."

        logger.debug(f'Condensed observations from: {on} to {maxN}')
        ds = ds.isel({self.obsdim: slice(0, maxN)})

        # Write new observation coordinate.
        obs = np.arange(0, maxN)
        ds = ds.assign_coords({self.obsdim: obs})

        return ds

    @__require_obsdim__
    def gridtime(self, times, timedim=None, round=True):
        if isinstance(times, str) or isinstance(times, pd.Timedelta):  # Make time series with given interval
            if round is True:
                start_time = np.nanmin(np.asarray(self.ds.time.dt.floor(times)))
                end_time = np.nanmax(np.asarray(self.ds.time.dt.ceil(times)))
            else:
                start_time = np.nanmin(np.asarray(self.ds.time))
                end_time = np.nanmax(np.asarray(self.ds.time))
            times = pd.date_range(start_time,
                                  end_time,
                                  freq=times,
                                  inclusive='both')

        if not isinstance(times, np.ndarray):
            times = times.to_numpy()

        timedim = self.timedim if timedim is None else timedim

        d = None

        for t in range(self.ds.sizes['trajectory']):
            dt = self.ds.isel(trajectory=t) \
                        .dropna(self.obsdim, how='all')

            dt = dt.assign_coords({self.obsdim : dt[self.timedim].values }) \
                   .drop_vars(self.timedim) \
                   .rename({self.obsdim : timedim}) \
                   .set_index({timedim: timedim})

            _, ui = np.unique(dt[timedim], return_index=True)
            dt = dt.isel({timedim: ui})
            dt = dt.isel({timedim : np.where(~pd.isna(dt[timedim].values))[0]})

            if dt.sizes[timedim] > 0:
                dt = dt.interp({timedim: times})
            else:
                logger.warning(f"time dimension ({timedim}) is zero size")

            if d is None:
                d = dt.expand_dims('trajectory')
            else:
                d = xr.concat((d, dt), "trajectory")

        d = d.assign_coords({'trajectory': self.ds['trajectory']})

        return d



from pathlib import Path
import xarray as xr
import logging
import pandas as pd
from dataclasses import dataclass
import numpy as np
import datetime

from .omb_decoder import decode_message
from typing import Union
from .omb_decoder import GNSS_Metadata, Waves_Metadata, Thermistors_Metadata, GNSS_Packet, Waves_Packet, Thermistors_Packet, _BD_YWAVE_NBR_BINS

logger = logging.getLogger(__name__)


@dataclass
class ParsedIridiumMessage:
    device_from: str
    kind: str
    meta: Union[GNSS_Metadata, Waves_Metadata, Thermistors_Metadata]
    data: Union[GNSS_Packet, Waves_Packet, Thermistors_Packet]


def sliding_filter_nsigma(np_array_in, nsigma=5.0, side_half_width=2):
    """Perform a sliding filter, on points of indexes
    [idx-side_half_width; idx+side_half_width], to remove outliers. I.e.,
    the [idx] point gets removed if it is more than nsigma deviations away
    from the mean of the whole segment.

    np_array_in should have a shape (nbr_of_entries,)"""

    np_array = np.copy(np_array_in)
    array_len = np_array.shape[0]

    middle_point_index_start = side_half_width
    middle_point_index_end = array_len - side_half_width - 1

    for crrt_middle_index in range(middle_point_index_start,
                                   middle_point_index_end + 1, 1):
        crrt_left_included = crrt_middle_index - side_half_width
        crrt_right_included = crrt_middle_index + side_half_width
        crrt_array_data = np.concatenate([
            np_array_in[crrt_left_included:crrt_middle_index],
            np_array_in[crrt_middle_index + 1:crrt_right_included + 1]
        ])
        mean = np.mean(crrt_array_data)
        std = np.std(crrt_array_data)
        if np.abs(np_array[crrt_middle_index] - mean) > nsigma * std:
            logger.debug("found outlier in sliding_filter_nsigma")
            np_array[crrt_middle_index] = np.nan

    return np_array


def append_dict_with_entry(dict_in: dict, parsed_entry: ParsedIridiumMessage):
    if parsed_entry.device_from not in dict_in:
        dict_in[parsed_entry.device_from] = {}
        dict_in[parsed_entry.device_from]["G"] = []
        dict_in[parsed_entry.device_from]["Y"] = []
        dict_in[parsed_entry.device_from]["T"] = []
    dict_in[parsed_entry.device_from][parsed_entry.kind].append(parsed_entry)


def read_omb_csv(path_in: Path,
                 dict_instruments_params: dict = None,
                 modified_wave_packet_properties: dict = None) -> xr.Dataset:
    logger.debug("read path to pandas")

    if modified_wave_packet_properties is None:
        nbr_bins_waves = _BD_YWAVE_NBR_BINS
    else:
        nbr_bins_waves = modified_wave_packet_properties[
            "_BD_YWAVE_PACKET_MAX_BIN"] - modified_wave_packet_properties[
                "_BD_YWAVE_PACKET_MIN_BIN"]

    ########################################
    # generic pandas read to be able to open from a variety of files

    omb_dataframe = pd.read_csv(path_in)

    ########################################
    # check this is actually a Rock7 data file

    columns = omb_dataframe.columns.to_list()
    expected_columns = [
        'Date Time (UTC)', 'Device', 'Direction', 'Payload', 'Approx Lat/Lng',
        'Payload (Text)', 'Length (Bytes)', 'Credits'
    ]

    if not set(expected_columns).issubset(set(columns)):
        raise RuntimeError(
            f"does not look like a Rock7 file; got colmns {columns}, expected {expected_columns}, missing: {set(expected_columns) - set(columns)}"
        )

    ########################################
    # decode

    dict_entries = {}
    number_valid_entries = 0

    frequencies = nbr_bins_waves * [np.nan]
    frequencies_set = False

    # decode each entry;
    # only consider messages from the buoy to owner
    # only consider non empty messages
    # be verbose about messages that cannot be decoded: something is seriously wrong then!
    for pd_index, crrt_data in omb_dataframe.iterrows():
        # only use data from the buoy
        if crrt_data.Direction != "MO":
            logger.debug(
                f"omb_dataframe at index {pd_index} is:\n{crrt_data}\nthis is not a from buoy (Direction: MO) message, drop"
            )
            continue

        # only use non empty data frames
        if getattr(crrt_data, "Length (Bytes)") == 0:
            logger.debug(
                f"omb_dataframe at index {pd_index} is:\n{crrt_data}\nthis is empty (Length (Bytes) is 0), drop"
            )
            continue

        # only use data that are after the start time for the current buoy
        crrt_start_time = None

        if dict_instruments_params is not None:
            crrt_instrument = getattr(crrt_data, "Device")
            if crrt_instrument in dict_instruments_params:
                if "start_time" in dict_instruments_params[crrt_instrument]:
                    crrt_start_time = dict_instruments_params[crrt_instrument]["start_time"]

        # we should only have valid dataframes at this point; attempt to decode
        # hard to catch exceptions in a fine grain way, so cath all, but only on the decoding itself
        # however, in practice, all messages should be decodable, and if not this is a serious issue; don t be silent
        try:
            crrt_kind, crrt_meta, crrt_list_packets = decode_message(
                crrt_data.Payload,
                print_decoded=False,
                dict_wave_packet_params=modified_wave_packet_properties)

        except AssertionError as e:
            logger.warning(
                f"attempt to decode entry at index {pd_index}, Payload equal to: {crrt_data.Payload} failed with exception:\n{e}"
            )
            continue

        number_valid_entries += 1

        # a GNSS packet may contain several data entries; split it here for simplicity
        if crrt_kind == "G":
            for crrt_fix in crrt_list_packets:
                if crrt_start_time is None or crrt_fix.datetime_fix > crrt_start_time:
                    crrt_parsed = ParsedIridiumMessage(
                        device_from=crrt_data.Device,
                        kind=crrt_kind,
                        meta=crrt_meta,
                        data=crrt_fix,
                    )

                    append_dict_with_entry(dict_entries, crrt_parsed)

                else:
                    logger.info(f"buoy {crrt_instrument}: ignore fix {crrt_fix}, since before {crrt_start_time}")

        # other packets contain a single entry: add as is
        else:
            if crrt_kind == "Y" and not frequencies_set:
                frequencies_set = True
                frequencies = crrt_list_packets[0].list_frequencies

            if crrt_start_time is None or crrt_list_packets[0].datetime_fix > crrt_start_time:
                crrt_parsed = ParsedIridiumMessage(
                    device_from=crrt_data.Device,
                    kind=crrt_kind,
                    meta=crrt_meta,
                    data=crrt_list_packets[0],
                )

                append_dict_with_entry(dict_entries, crrt_parsed)

            else:
                logger.info(f"buoy {crrt_instrument}: ignore spectrum with timestamp {crrt_list_packets[0].datetime_fix}, since before {crrt_start_time}")

    if number_valid_entries == 0:
        logger.warning(
            "got no valid decoded payload in the whole csv file; are you sure this is an OMB csv iridium file?"
        )

    ########################################
    # turn into a trajan compatible format

    # determine the size for trajectory, obs, imu_obs, frequencies
    trajectory = len(dict_entries)

    obs_gnss = max(
        [len(dict_entries[crrt_instr]["G"]) for crrt_instr in dict_entries])

    obs_waves_imu = max(
        [len(dict_entries[crrt_instr]["Y"]) for crrt_instr in dict_entries])

    frequencies_waves_imu = nbr_bins_waves

    list_instruments = sorted(list(dict_entries.keys()))

    empty_time = np.full((trajectory, obs_gnss),
                         np.datetime64('nat'),
                         dtype='datetime64[ns]')
    empty_time_waves_imu = np.full((trajectory, obs_waves_imu),
                                   np.datetime64('nat'),
                                   dtype='datetime64[ns]')

    # create and fill the xarray dataset
    xr_result = xr.Dataset(
        {
            # meta vars
            #
            'trajectory':
            xr.DataArray(data=list_instruments,
                         dims=['trajectory'],
                         attrs={
                             "cf_role": "trajectory_id",
                             "standard_name": "platform_id",
                         }).astype(str),
            #
            'frequencies_waves_imu':
            xr.DataArray(data=frequencies,
                         dims=["frequencies_waves_imu"],
                         attrs={
                             "_FillValue": "NaN",
                             "unit": "Hz",
                         }),
            #
            # gnss position vars
            #
            'time':
            xr.DataArray(dims=["trajectory", "obs"],
                         data=empty_time,
                         attrs={
                             "standard_name": "time",
                         }),
            #
            'lat':
            xr.DataArray(dims=["trajectory", "obs"],
                         data=np.nan * np.ones((trajectory, obs_gnss)),
                         attrs={
                             "_FillValue": "NaN",
                             "standard_name": "latitude",
                             "unit": "degree_north",
                         }),
            #
            'lon':
            xr.DataArray(dims=["trajectory", "obs"],
                         data=np.nan * np.ones((trajectory, obs_gnss)),
                         attrs={
                             "_FillValue": "NaN",
                             "standard_name": "longitude",
                             "unit": "degree_east",
                         }),
            #
            # imu waves vars
            #
            'time_waves_imu':
            xr.DataArray(dims=["trajectory", "obs_waves_imu"],
                         data=empty_time_waves_imu,
                         attrs={
                             "standard_name": "time",
                         }),
            #
            'accel_energy_spectrum':
            xr.DataArray(
                dims=["trajectory", "obs_waves_imu", "frequencies_waves_imu"],
                data=np.nan * np.ones(
                    (trajectory, obs_waves_imu, frequencies_waves_imu)),
                attrs={
                    "_FillValue": "NaN",
                }),
            #
            'elevation_energy_spectrum':
            xr.DataArray(
                dims=["trajectory", "obs_waves_imu", "frequencies_waves_imu"],
                data=np.nan * np.ones(
                    (trajectory, obs_waves_imu, frequencies_waves_imu)),
                attrs={
                    "_FillValue": "NaN",
                }),
            #
            'processed_elevation_energy_spectrum':
            xr.DataArray(
                dims=["trajectory", "obs_waves_imu", "frequencies_waves_imu"],
                data=np.nan * np.ones(
                    (trajectory, obs_waves_imu, frequencies_waves_imu)),
                attrs={
                    "_FillValue": "NaN",
                }),
            #
            'pcutoff':
            xr.DataArray(dims=["trajectory", "obs_waves_imu"],
                         data=np.nan * np.ones((trajectory, obs_waves_imu)),
                         attrs={
                             "_FillValue": "NaN",
                         }),
            #
            'pHs0':
            xr.DataArray(
                dims=["trajectory", "obs_waves_imu"],
                data=np.nan * np.ones((trajectory, obs_waves_imu)),
                attrs={
                    "_FillValue":
                    "NaN",
                    "definition":
                    "4 * math.sqrt(m0) of low freq cutoff elevation spectrum"
                }),
            #
            'pT02':
            xr.DataArray(
                dims=["trajectory", "obs_waves_imu"],
                data=np.nan * np.ones((trajectory, obs_waves_imu)),
                attrs={
                    "_FillValue":
                    "NaN",
                    "definition":
                    "math.sqrt(m0 / m2) of low freq cutoff elevation spectrum"
                }),
            #
            'pT24':
            xr.DataArray(
                dims=["trajectory", "obs_waves_imu"],
                data=np.nan * np.ones((trajectory, obs_waves_imu)),
                attrs={
                    "_FillValue":
                    "NaN",
                    "definition":
                    "math.sqrt(m2 / m4) of low freq cutoff elevation spectrum"
                }),
            #
            'Hs0':
            xr.DataArray(dims=["trajectory", "obs_waves_imu"],
                         data=np.nan * np.ones((trajectory, obs_waves_imu)),
                         attrs={
                             "_FillValue":
                             "NaN",
                             "definition":
                             "4 * math.sqrt(m0) of full elevation spectrum"
                         }),
            #
            'T02':
            xr.DataArray(dims=["trajectory", "obs_waves_imu"],
                         data=np.nan * np.ones((trajectory, obs_waves_imu)),
                         attrs={
                             "_FillValue":
                             "NaN",
                             "definition":
                             "math.sqrt(m0 / m2) of full elevation spectrum"
                         }),
            #
            'T24':
            xr.DataArray(dims=["trajectory", "obs_waves_imu"],
                         data=np.nan * np.ones((trajectory, obs_waves_imu)),
                         attrs={
                             "_FillValue":
                             "NaN",
                             "definition":
                             "math.sqrt(m2 / m4) of full elevation spectrum"
                         }),
        }, )

    # actually fill the data
    for crrt_instrument_idx, crrt_instrument in enumerate(list_instruments):
        ####################
        # gnss position data

        list_time = [
            crrt_packet.data.datetime_posix
            for crrt_packet in dict_entries[crrt_instrument]["G"]
        ]
        list_lat = [
            crrt_packet.data.latitude
            for crrt_packet in dict_entries[crrt_instrument]["G"]
        ]
        list_lon = [
            crrt_packet.data.longitude
            for crrt_packet in dict_entries[crrt_instrument]["G"]
        ]

        # sort in time
        argsort_time_idx = list(np.argsort(np.array(list_time)))
        list_time = [list_time[i] for i in argsort_time_idx]
        list_lat = [list_lat[i] for i in argsort_time_idx]
        list_lon = [list_lon[i] for i in argsort_time_idx]

        # reject outliers
        logger.debug("start applying sliding_filter_nsigma")
        np_latitude = sliding_filter_nsigma(np.array(list_lat))
        np_longitude = sliding_filter_nsigma(np.array(list_lon))
        logger.debug("done applying sliding_filter_nsigma")

        xr_result["time"][crrt_instrument_idx,
                          0:len(list_time)] = pd.to_datetime(list_time,
                                                             utc=True,
                                                             unit='s')
        xr_result["lat"][crrt_instrument_idx, 0:len(list_lat)] = np_latitude
        xr_result["lon"][crrt_instrument_idx, 0:len(list_lon)] = np_longitude

        ####################
        # wave data

        list_parsed_waves_messages = dict_entries[crrt_instrument]["Y"]

        crrt_list_times_waves = [
            crrt_wave_data.data.datetime_posix
            for crrt_wave_data in list_parsed_waves_messages
        ]
        argsort_time_idx = list(np.argsort(np.array(crrt_list_times_waves)))
        list_parsed_waves_messages = [
            list_parsed_waves_messages[i] for i in argsort_time_idx
        ]

        for crrt_wave_idx, crrt_wave_data in enumerate(
                list_parsed_waves_messages):
            xr_result["time_waves_imu"][crrt_instrument_idx, crrt_wave_idx] = \
                pd.to_datetime(crrt_wave_data.data.datetime_posix, utc=True, unit='s')

            xr_result["pcutoff"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.low_frequency_index_cutoff

            xr_result["accel_energy_spectrum"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.list_acceleration_energies

            xr_result["elevation_energy_spectrum"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.list_elevation_energies

            xr_result["processed_elevation_energy_spectrum"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.processed_list_elevation_energies

            xr_result["pHs0"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.processed_Hs

            xr_result["pT02"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.processed_Tz

            xr_result["pT24"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.processed_Tc

            xr_result["Hs0"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.Hs

            xr_result["T02"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.Tz

            xr_result["T24"][crrt_instrument_idx, crrt_wave_idx] = \
                crrt_wave_data.data.Tc

    xr_result = xr_result.traj.assign_cf_attrs(
        creator_name="XX:TODO",
        creator_email="XX:TODO",
        title="XX:TODO",
        summary="XX:TODO",
        history=
        "created with trajan.reader.omb from a Rock7 Iridium CSV file of OMB transmissions"
    )

    return xr_result



# Utility script to quickly plot a drifter collection file

import numpy as np
import matplotlib.pyplot as plt
import xarray as xr
import trajan as ta
import click
from pathlib import Path
import lzma

@click.command()
@click.argument('tf')
@click.option('-l', '--land',
              default='auto',
              help='Scale of coastline (f, h, i, l, c or mask)',
              type=str,
              multiple=False)
@click.option('-m', '--margin',
              default=.1,
              help='Margin/whitespace in degrees around drifter coverage',
              type=float,
              multiple=False)
@click.option('-s', '--start_time',
              default=None,
              help='Show only data after this time',
              type=str,
              multiple=False)
@click.option('-e', '--end_time',
              default=None,
              help='Show only data before this time',
              type=str,
              multiple=False)

def main(tf, land, start_time, end_time, margin):
    tf = Path(tf)
    if tf.suffix == '.xz':
        with lzma.open(tf) as fd:
            ds = xr.open_dataset(fd)
            ds.load()
    else:
        ds = xr.open_dataset(tf)

    if 'status' in ds:  # hack for OpenDrift files
        ds = ds.where(ds.status>=0)

    trajectory_names = None
    color = 'gray'
    dsub = ds.filter_by_attrs(cf_role='trajectory_id')
    if len(dsub.data_vars) == 1:
        trajectory_names = dsub.to_array().values[0]
        color = None

    if start_time is not None:
        ds = ds.where(ds.time>np.datetime64(start_time))
    if end_time is not None:
        ds = ds.where(ds.time<np.datetime64(end_time))

    ds.traj.plot(label=trajectory_names, color=color, land=land, margin=margin)

    start_time = np.nanmin(ds.time.data).astype('datetime64[s]')
    end_time = np.nanmax(ds.time.data).astype('datetime64[s]')
    name = tf

    plt.gca().set_title(f'{name} [ {start_time} to {end_time} ]')
    plt.legend()

    plt.show()

if __name__ == '__main__':
    main()



from functools import partial
import logging
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import cartopy.crs as ccrs
import numpy as np
import xarray as xr

from trajan.plot import Plot

logger = logging.getLogger(__name__)
logging.getLogger('matplotlib.font_manager').disabled = True

class Animation:
    ds: xr.Dataset

    # A lon-lat projection with the currently used globe.
    gcrs = None

    DEFAULT_LINE_COLOR = Plot.DEFAULT_LINE_COLOR

    def __init__(self, ds):
        self.ds = ds
        self.gcrs = ccrs.PlateCarree()

    def __call__(self, *args, **kwargs):
        return self.animate(*args, **kwargs)

    def animate(self, *args, **kwargs):
        logger.debug(f'Animating trajectories..')
        ax = self.ds.traj.plot.set_up_map(kwargs)
        fig = ax.get_figure()

        ds = self.ds.traj.gridtime(times='1h')
        frames = ds.time.shape[0]

        logger.debug(f'Running animation.. frames: {frames}')
        self.__animation__ = FuncAnimation(fig, partial(self.plot_frame, ax, ds), frames=ds.time.shape[0], interval=20)

        return self.__animation__

    def plot_frame(self, ax, ds, frame):
        logger.debug(f'Plotting frame: {frame}')
        ds = ds.isel(time=frame)
        x = ds.lon.values.T
        y = ds.lat.values.T
        paths = ax.scatter(x, y, transform=self.gcrs)
        return paths





import logging
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import numpy as np
import xarray as xr

from .land import add_land

logger = logging.getLogger(__name__)
logging.getLogger('matplotlib.font_manager').disabled = True


class Plot:
    ds: xr.Dataset

    # A lon-lat projection with the currently used globe.
    gcrs = None

    DEFAULT_LINE_COLOR = 'gray'

    def __init__(self, ds):
        self.ds = ds
        self.gcrs = ccrs.PlateCarree()

    @property
    def __cartesian__(self):
        return self.ds.traj.crs is None

    def set_up_map(
        self,
        kwargs_d=None,
        **kwargs
    ):
        """
        Set up axes for plotting.

        Args:

            crs: Use a different crs than Mercator.

            margin: margin (decimal degrees) in addition to extent of trajectories.

            land: Add land shapes based on GSHHG to map.

                'auto' (default): use automatic scaling.

                'c', 'l','i','h','f' or
                'coarse', 'low', 'intermediate', 'high', 'full': use corresponding GSHHG level.

                'mask' or 'fast' (fastest): use a raster mask generated from GSHHG.

                None: do not add land shapes.

        Returns:

            An matplotlib axes with a Cartopy projection.

        """
        if self.__cartesian__:
            ax = plt.axes()
            return ax

        # By popping the args from kwargs they are not passed onto matplotlib later.
        if kwargs_d is None:
            kwargs_d = kwargs
        else:
            for k,v in kwargs:
                kwargs_d[k] = v

        ax = kwargs_d.pop('ax', None)
        crs = kwargs_d.pop('crs', None)
        margin = kwargs_d.pop('margin', .1)
        corners = kwargs_d.pop('corners', None)
        land = kwargs_d.pop('land', 'auto')
        figsize = kwargs_d.pop('figsize', 11)

        assert crs is None or ax is None, "Only one of `ax` and `crs` may be specified."

        if ax is not None:
            logger.debug('axes already set up')
            return ax

        # It is not possible to change the projection of existing axes. The type of axes object returned
        # by `plt.axes` depends on the input projection.

        # Create a new figure if none exists.
        if corners is None:
            lonmin = self.ds.lon.min() - margin
            lonmax = self.ds.lon.max() + margin
            latmin = self.ds.lat.min() - margin
            latmax = self.ds.lat.max() + margin
        else:
            lonmin = corners[0]
            lonmax = corners[1]
            latmin = corners[2]
            latmax = corners[3]

        if len(plt.get_fignums()) == 0:
            logger.debug('Creating new figure and axes..')
            meanlat = (latmin + latmax) / 2
            aspect_ratio = float(latmax - latmin) / (float(lonmax - lonmin))
            aspect_ratio = np.abs(aspect_ratio / np.cos(np.radians(meanlat)))

            if aspect_ratio > 1:
                fig = plt.figure(figsize=(figsize / aspect_ratio, figsize))
            else:
                fig = plt.figure(figsize=(figsize, figsize * aspect_ratio))

            # fig.canvas.draw()  # maybe needed?
            plt.tight_layout()

        else:
            fig = plt.gcf()
            if len(fig.axes) > 0:
                logger.debug('Axes already exist on existing figure.')
                return fig.gca()
            else:
                logger.debug('Figure exists, setting up axes.')

        crs = crs if crs is not None else ccrs.Mercator()
        self.gcrs = ccrs.PlateCarree(globe=crs.globe)

        ax = fig.add_subplot(111, projection=crs)
        ax.set_extent([lonmin, lonmax, latmin, latmax], crs=self.gcrs)

        gl = ax.gridlines(self.gcrs, draw_labels=True)
        gl.top_labels = None

        if land is not None:
            add_land(ax, lonmin, latmin, lonmax, latmax,
                 fast=(land == 'mask' or land == 'fast'), lscale=land, globe=crs.globe)

        return ax

    def __call__(self, *args, **kwargs):
        return self.lines(*args, **kwargs)

    def lines(self, *args, **kwargs):
        """
        Plot the trajectory lines.

        Args:

            ax: Use existing axes, otherwise a new one is set up.

            crs: Specify crs for new axis.

        Returns:

            Matplotlib lines, and axes.
        """
        logger.debug(f'Plotting lines')
        ax = self.set_up_map(kwargs)

        if 'color' not in kwargs:
            kwargs['color'] = self.DEFAULT_LINE_COLOR

        if 'alpha' not in kwargs and 'trajectory' in self.ds.dims:
            num = self.ds.sizes['trajectory']
            if num>100:  # If many trajectories, make more transparent
                kwargs['alpha'] = np.maximum(.1, 100/np.float64(num))

        if self.__cartesian__:
            x = self.ds.traj.tx.values.T
            y = self.ds.traj.ty.values.T
        else:
            x = self.ds.traj.tlon.values.T
            y = self.ds.traj.tlat.values.T

        if hasattr(kwargs['color'], 'shape'):
            from matplotlib.collections import LineCollection
            c = kwargs.pop('color').T
            if hasattr(c, 'values'):
                c = c.values
            vmin = kwargs.pop('vmin', np.nanmin(c))
            vmax = kwargs.pop('vmax', np.nanmax(c))
            norm = plt.Normalize(vmin, vmax)
            colorbar = kwargs.pop('colorbar', False)

            for i in range(x.shape[1]):
                logger.debug(f'Plotting trajectory {i} of {x.shape[1]} with color')
                points = np.array([x[:,i].T, y[:,i].T]).T.reshape(-1, 1, 2)
                segments = np.concatenate([points[:-1], points[1:]], axis=1)
                if self.__cartesian__:
                    lc = LineCollection(segments, cmap='jet', norm=norm,
                                        *args, **kwargs)
                else:
                    lc = LineCollection(segments, cmap='jet', norm=norm, transform=self.gcrs,
                                        *args, **kwargs)
                # Set the values used for colormapping
                lc.set_array(c[:,i])
                paths = ax.add_collection(lc)
        else:
            if self.__cartesian__:
                paths = ax.plot(x, y, *args, **kwargs)
            else:
                paths = ax.plot(x, y, transform=self.gcrs, *args, **kwargs)

        return paths

    def scatter(self, *args, **kwargs):
        """
        Plot the particles as points.

        Args:

            ax: Use existing axes, otherwise a new one is set up.

            crs: Specify crs for new axis.

        Returns:

            Matplotlib lines, and axes.
        """

        logger.debug(f'Plotting points')
        ax = self.set_up_map(kwargs)

        if 'marker' not in kwargs:
            kwargs['marker'] = '.'

        if 'color' not in kwargs:
            kwargs['color'] = self.DEFAULT_LINE_COLOR

        if 'alpha' not in kwargs and 'trajectory' in self.ds.dims:
            num = self.ds.sizes['trajectory']
            if num>100:  # If many particles, make more transparent
                kwargs['alpha'] = np.maximum(.1, 100/np.float64(num))

        if self.__cartesian__:
            x = self.ds.traj.tx.values.T
            y = self.ds.traj.ty.values.T
        else:
            x = self.ds.traj.tlon.values.T
            y = self.ds.traj.tlat.values.T

        if self.__cartesian__:
            paths = ax.scatter(x, y, *args, **kwargs)
        else:
            paths = ax.scatter(x, y, transform=self.gcrs, *args, **kwargs)

        return paths

    def convex_hull(self, *args, **kwargs):
        """
        Plot the convex hull around all particles

        Args:

            ax: Use existing axes, otherwise a new one is set up.

            crs: Specify crs for new axis.

        Returns:

            Matplotlib lines, and axes.
        """

        logger.debug(f'Plotting convex hull')
        hull = self.ds.traj.convex_hull()

        ax = self.set_up_map(kwargs)

        if 'color' not in kwargs:
            kwargs['color'] = self.DEFAULT_LINE_COLOR

        # TODO: might not work for cartesian plots
        line_segments = [hull.points[simplex] for simplex in hull.simplices]
        from matplotlib.collections import LineCollection
        paths = ax.add_collection(LineCollection(line_segments, transform=self.gcrs,
                                                 *args, **kwargs))

        return paths




"""
Plot land shapes and landmask.
"""
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import numpy as np
import logging

logger = logging.getLogger(__name__)

__mask__ = None


def __get_mask__():
    """
    Return an instance of the global landmask.
    """
    global __mask__

    try:
        # Use the cached landmask from OpenDrift if it exists
        from opendrift.readers import reader_global_landmask
        if reader_global_landmask.__roaring_mask__ is not None:
            logger.debug('Using roaring landmask instance from OpenDrift.')
            __mask__ = reader_global_landmask.__roaring_mask__
    except ImportError:
        pass

    if __mask__ is None:
        logger.debug('Setting up roaring landmask..')
        from roaring_landmask import RoaringLandmask
        __mask__ = RoaringLandmask.new()

    return __mask__


def add_land(ax,
             lonmin,
             latmin,
             lonmax,
             latmax,
             fast,
             ocean_color='white',
             land_color=cfeature.COLORS['land'],
             lscale='auto',
             globe=None):
    """
    Plot the landmask or the shapes from GSHHG.
    """

    def show_landmask_roaring(roaring):
        maxn = 512.
        dx = (lonmax - lonmin) / maxn
        dy = (latmax - latmin) / maxn
        dx = max(roaring.dx, dx)
        dy = max(roaring.dy, dy)

        x = np.arange(lonmin, lonmax, dx)
        y = np.arange(latmin, latmax, dy)

        yy, xx = np.meshgrid(y, x)
        img = roaring.mask.contains_many_par(xx.ravel(),
                                             yy.ravel()).reshape(yy.shape).T

        from matplotlib import colors
        cmap = colors.ListedColormap([ocean_color, land_color])
        ax.imshow(img,
                  origin='lower',
                  extent=[lonmin, lonmax, latmin, latmax],
                  zorder=0,
                  transform=ccrs.PlateCarree(globe=globe),
                  cmap=cmap)

    if fast:
        show_landmask_roaring(__get_mask__())
    else:
        land = cfeature.GSHHSFeature(scale=lscale, facecolor=land_color)

        ax.add_feature(land, zorder=2, facecolor=land_color, edgecolor='black')

